;; -*- lisp -*-

;;;; ** Generic cookie management.

;;;; Backends which can provide more efficent versions of these
;;;; methods are strongly advised to do so.

(in-package :it.bese.ucw)

(defclass cookie ()
  ((value :accessor value :initarg :value :initform nil)
   (path :accessor path :initarg :path :initform nil)
   (expires :accessor expries :initarg :expires :initform nil)
   (domain :accessor domain :initarg :domain :initform nil)
   (secure :accessor secure :initarg :secure :initform nil)))

(defgeneric cookie (request cookie-id)
  (:documentation "Returns the value associated with the cookie
  with id COOKIE-ID."))

(defun parse-cookie-string (cookie cookie-id)
  (ucw.backend.dribble "Parsing cookie ~S." cookie)
  (loop with name-buffer = (make-array (length cookie) :element-type 'character
                                       :adjustable t
                                       :fill-pointer 0)
     with value-buffer = (make-array (length cookie) :element-type 'character
                                     :adjustable t
                                     :fill-pointer 0)
     with state = :name
     for char across cookie
     do (ecase state
          (:name (if (char= char #\=)
                     (setf state :value)
                     (vector-push-extend char name-buffer)))
          (:value (if (member char '(#\; #\Space #\Tab))
                      (if (string= cookie-id name-buffer)
                          ;; if the name we've collected is ours
                          ;; then we're done.
                          (return-from parse-cookie-string (copy-seq value-buffer))
                          ;; otherwise we need to keep searching
                          (setf state :none
                                (fill-pointer name-buffer) 0
                                (fill-pointer value-buffer) 0))
                      (vector-push-extend char value-buffer)))
          (:none (unless (member char '(#\; #\Space #\Tab))
                   (vector-push-extend char name-buffer)
                   (setf state :name))))
     finally (return
               ;; be liniente if there's no trailing end of value marker
               (if (string= cookie-id name-buffer)
                   ;; if the name we've collected is ours
                   ;; then we're done.
                   (prog1
                       (copy-seq value-buffer)
                     (ucw.backend.dribble "Parsed ~S into ~S." cookie value-buffer))
                   ;; nope, didn't find it
                   nil))))

(defmethod cookie ((message request) cookie-id)
  (parse-cookie-string (or (get-header message "Cookie") "") cookie-id))

(defgeneric (setf cookie) (new-cookie message cookie-id))

(defmethod (setf cookie) (new-cookie (res response) name)
  (add-header res
              "Set-Cookie" (with-output-to-string (cookie)
			     (with-slots (value path expires domain secure)
				 new-cookie
			       (format cookie "~A=~A;" name value)
			       (when path    (format cookie " path=~A;" path))
			       (when expires (format cookie " expires=~A;" expires))
			       (when domain  (format cookie " domain=~A;" domain))
			       (when secure  (format cookie " secure"))))))

;; Copyright (c) 2003-2005 Edward Marco Baringer
;; All rights reserved. 
;; 
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;; 
;;  - Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;; 
;;  - Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;; 
;;  - Neither the name of Edward Marco Baringer, nor BESE, nor the names
;;    of its contributors may be used to endorse or promote products
;;    derived from this software without specific prior written permission.
;; 
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
